package MPI ;

public class Datatype {
  public final static int NULL    = 0;
  public final static int BYTE    = 1;
  public final static int CHAR    = 2;
  public final static int SHORT   = 3;
  public final static int BOOLEAN = 4;
  public final static int INT     = 5;
  public final static int LONG    = 6;
  public final static int FLOAT   = 7;
  public final static int DOUBLE  = 8;  
  public final static int PACKED  = 9;
  public final static int LB      =10;
  public final static int UB      =11;

  public Datatype(int Type) 
  { handle = GetDatatype(Type); baseType = this;}
  public Datatype(int count, int[] blocklengths,   
    int[] displacements, Datatype[] types) { 
    handle = GetStruct(count, blocklengths, displacements, types);
    baseType = types[0];
  }

  public native int Extent();
  public native int Size();
  public native int Lb();
  public native int Ub();
  public native void Commit();
  public void finalize() { Finalize();}

  private native void Finalize();

  public int baseExtent() { return baseType.Extent();}
  public int baseSize() { return baseType.Size();}

  public Datatype Contiguous(int count) 
  { return new Datatype(GetContiguous(count), this);}
  public Datatype Vector(int count, int blocklength, int stride)
  { return new Datatype(GetVector(count, blocklength, stride), this);}
  public Datatype Hvector(int count, int blocklength, int stride)
  { return new Datatype(GetHvector(count, blocklength, stride), this);}
  public Datatype Indexed(int count, int[] blocklengths, 
    int[] displacements)
  { return new Datatype(GetIndexed(count, blocklengths,
    displacements), this);}
  public Datatype Hindexed(int count, int[] blocklengths, 
    int[] displacements)
  { return new Datatype(GetHindexed(count, blocklengths,
    displacements), this);}

  private Datatype(long _handle, Datatype base) 
  { handle = _handle; baseType = base;}
  private native long GetDatatype(int Type);
  private native long GetContiguous(int count);
  private native long GetVector(int count, int blocklength, int stride);
  private native long GetHvector(int count, int blocklength, int stride);
  private native long GetIndexed(int count, int[] blocklengths, 
    int[] displacements);
  private native long GetHindexed(int count, int[] blocklengths, 
    int[] displacements);  
  private native long GetStruct(int count, int[] blocklengths, 
    int[] displacements, Datatype[] types);  

  protected long handle;
  protected Datatype baseType;
}

