#include "MPI_Datatype.h"
#include "MPI_Comm.h"
#include "MPI_Cart.h"
#include <mpi.h>

#define COMM(x) ((MPI_Comm)*(void**)&(unhand(x)->handle))
#define DATATYPE(x) ((MPI_Datatype)*(void**)&(unhand(x)->handle))
#define REF(x) (*(void**)&(x))

void MPI_Cart_Get(struct HMPI_Cart *this,
  long maxdims, HArrayOfInt *dims,
  HArrayOfInt *periods, HArrayOfInt *coords) {
  int i;
  int *ds = (int*)calloc(maxdims, sizeof(int));
  int *ps = (int*)calloc(maxdims, sizeof(int));
  int *cs = (int*)calloc(maxdims, sizeof(int));
  MPI_Cart_get(COMM(this), maxdims, ds, ps, cs);
  for (i=0; i<maxdims; i++) {
    unhand(dims)->body[i] = ds[i];
    unhand(periods)->body[i] = ps[i];
    unhand(coords)->body[i] = cs[i];
  }
  free(ds);
  free(ps);
  free(cs);
}

long MPI_Cart_Rank(struct HMPI_Cart *this,
  HArrayOfInt *coords) {
  int i, result;
  int maxdims = obj_length(coords);
  int *cs = (int*)calloc(maxdims, sizeof(int));
  MPI_Cart_rank(COMM(this), cs, &result);
  for (i=0; i<maxdims; i++)
    unhand(coords)->body[i] = cs[i];
  free(cs);
  return result;
}

void MPI_Cart_Coords(struct HMPI_Cart *this,
  long rank, long maxdims, HArrayOfInt *coords)
{
  int i;
  int *cs = (int*)calloc(maxdims, sizeof(int));
  MPI_Cart_coords(COMM(this), rank, maxdims, cs);
  for (i=0; i<maxdims; i++) 
    unhand(coords)->body[i] = cs[i];
  free(cs);
}

void MPI_Cart_Shift(struct HMPI_Cart *this,
  long direction, long disp,
  HArrayOfInt *source,
  HArrayOfInt *dest) {
  int i;
  int maxdims = obj_length(source);
  int *ss = (int*)calloc(maxdims, sizeof(int));
  int *ds = (int*)calloc(maxdims, sizeof(int));
  MPI_Cart_shift(COMM(this), direction, disp, ss, ds);
  for (i=0; i<maxdims; i++) {
    unhand(source)->body[i] = ss[i];
    unhand(dest)->body[i] = ds[i];
  }
  free(ss);
  free(ds);
}

int64_t MPI_Cart_GetCart(struct HMPI_Cart *this,
  struct HMPI_Comm *old,
  long ndims,
  HArrayOfInt *dims,
  HArrayOfInt *periods,
  long reorder) {
  int i;
  int64_t result;
  MPI_Comm comm;
  int *ds = (int*)calloc(ndims, sizeof(int));
  int *ps = (int*)calloc(ndims, sizeof(int));
  for (i=0; i<ndims; i++) {
    ds[i] = unhand(dims)->body[i];
    ps[i] = unhand(periods)->body[i];
  }
  MPI_Cart_create(COMM(old), ndims, ds, ps, reorder, &comm);
  REF(result) = comm;
  return result;
}


