#include "MPI_Status.h"
#include "MPI_Datatype.h"
#include "MPI_Op.h"
#include "MPI_Comm.h"
#include <mpi.h>

#define COMM(x) ((MPI_Comm)*(void**)&(unhand(x)->handle))
#define OP(x) ((MPI_Op)*(void**)&(unhand(x)->handle))
#define DATATYPE(x) ((MPI_Datatype)*(void**)&(unhand(x)->handle))
#define REF(x) (*(void**)&(x))

int64_t MPI_Comm_GetComm(struct HMPI_Comm *this, long Type) {
  int64_t result;
  switch (Type) {
    case 0:
      REF(result) = 0;
      break;
    case 1:
      REF(result) = MPI_COMM_SELF;
      break;
    case 2:
      REF(result) = MPI_COMM_WORLD;
      break;
  }
  return result;
}
    
void MPI_Comm_Barrier(struct HMPI_Comm *this) {
  MPI_Barrier(COMM(this));
}

long MPI_Comm_Size(struct HMPI_Comm *this)
{
  int size;
  MPI_Comm_size(COMM(this), &size);
  return size;
}

long MPI_Comm_Rank(struct HMPI_Comm *this)
{
  int rank;
  MPI_Comm_rank(COMM(this), &rank);
  return rank;
}  

void MPI_Comm_Send(struct HMPI_Comm *this,
  struct Hjava_lang_Object *buf, 
  long offset, long baseSize, long count, 
  struct HMPI_Datatype *type,
  long dest, long tag) 
{
  MPI_Send(unhand((HArrayOfByte*)buf)->body+offset*baseSize, 
    count, DATATYPE(type), dest, tag, 
    COMM(this));
}

struct HMPI_Status* MPI_Comm_Recv(struct HMPI_Comm *this,
  struct Hjava_lang_Object *buf, 
  long offset, long count,
  struct HMPI_Datatype *type,
  long source, long tag,
  struct HMPI_Status *stat)
{
  MPI_Status status;
  MPI_Recv(unhand((HArrayOfByte*)buf)->body+offset, 
    count, DATATYPE(type), source, tag,
    COMM(this), &status);
  MPI_Get_count(&status, DATATYPE(type), 
    &status.count);
  unhand(stat)->source = status.MPI_SOURCE;
  unhand(stat)->tag = status.MPI_TAG;
  unhand(stat)->error = status.MPI_ERROR;
  unhand(stat)->count = status.count;
  return stat;
}

struct HMPI_Status* MPI_Comm_Sendrecv(struct HMPI_Comm *this,
  struct Hjava_lang_Object *sbuf, 
  long soffset, long scount,
  struct HMPI_Datatype *stype,
  long dest, long stag,
  struct Hjava_lang_Object *rbuf, 
  long roffset, long rcount,
  struct HMPI_Datatype *rtype,
  long source, long rtag,
  struct HMPI_Status *stat)
{
  MPI_Status status;
  MPI_Sendrecv(unhand((HArrayOfByte*)sbuf)->body+soffset, 
    scount, DATATYPE(stype), dest, stag,
    unhand((HArrayOfByte*)rbuf)->body+roffset, 
    rcount, DATATYPE(rtype), source, rtag,
    COMM(this), &status);
  MPI_Get_count(&status, DATATYPE(stype), 
    &status.count);
  unhand(stat)->source = status.MPI_SOURCE;
  unhand(stat)->tag = status.MPI_TAG;
  unhand(stat)->error = status.MPI_ERROR;
  unhand(stat)->count = status.count;
  return stat;
}


