#include <mpi.h>
#include "MPI_Datatype.h"

#define REF(x) (*(void**)&(x))
#define DATATYPE(x) ((MPI_Datatype)*(void**)&(unhand(x)->handle))
#define PDATATYPE(x) ((MPI_Datatype*)(void**)&(unhand(x)->handle))

long MPI_Datatype_Extent(struct HMPI_Datatype *this) {
  MPI_Aint result;
  MPI_Type_extent(DATATYPE(this), &result);
  return result;
}

long MPI_Datatype_Size(struct HMPI_Datatype *this) {
  int result;
  MPI_Type_size(DATATYPE(this), &result);
  return result;
}

long MPI_Datatype_Lb(struct HMPI_Datatype *this) {
  MPI_Aint result;
  MPI_Type_lb(DATATYPE(this), &result);
  return result;
}

long MPI_Datatype_Ub(struct HMPI_Datatype *this) {
  MPI_Aint result;
  MPI_Type_ub(DATATYPE(this), &result);
  return result;
}

void MPI_Datatype_Commit(struct HMPI_Datatype *this) {
  MPI_Type_commit(PDATATYPE(this));
}

void MPI_Datatype_Finalize(struct HMPI_Datatype *this) {
  MPI_Type_free(PDATATYPE(this));
  unhand(this)->handle = 0;
}


int64_t MPI_Datatype_GetDatatype(struct HMPI_Datatype *this, long Type) {
  int64_t result;
  static MPI_Datatype Dts[] = { 0, MPI_CHAR, MPI_SHORT, MPI_SHORT,
    MPI_LONG, MPI_LONG, MPI_LONG_LONG_INT, MPI_FLOAT, MPI_DOUBLE,
    MPI_PACKED, MPI_LB, MPI_UB};
  REF(result) = Dts[Type];
  return result;
}

int64_t MPI_Datatype_GetContiguous(struct HMPI_Datatype *this, long count)
{
  int64_t result;
  MPI_Datatype type;
  MPI_Type_contiguous(count, DATATYPE(this), &type);
  REF(result) = type;
  return result;
}

int64_t MPI_Datatype_GetVector(struct HMPI_Datatype *this, 
  long count, long blocklength, long stride) {
  int64_t result;
  MPI_Datatype type;
  MPI_Type_vector(count, blocklength, stride, DATATYPE(this), &type);
  REF(result) = type;
  return result;
}

int64_t MPI_Datatype_GetHvector(struct HMPI_Datatype *this,
  long count, long blocklength, long stride) {
  int64_t result;
  MPI_Datatype type;
  MPI_Type_hvector(count, blocklength, stride, DATATYPE(this), &type);
  REF(result) = type;
  return result;
}

int64_t MPI_Datatype_GetIndexed(struct HMPI_Datatype *this,
  long count, HArrayOfInt *blocklengths,
  HArrayOfInt *displacements) {
  int i;
  int64_t result;
  MPI_Datatype type;
  int *lengths = (int*)calloc(count, sizeof(int));
  int *disps = (int*)calloc(count, sizeof(int));
  for (i=0; i<count; i++) {
    lengths[i] = unhand(blocklengths)->body[i];
    disps[i] = unhand(displacements)->body[i];
  }
  MPI_Type_indexed(count, lengths, disps, DATATYPE(this), &type);
  REF(result) = type;
  free(lengths);
  free(disps);
  return result;
}

int64_t MPI_Datatype_GetHindexed(struct HMPI_Datatype *this,
  long count, HArrayOfInt *blocklengths,
  HArrayOfInt *displacements) {
  int i;
  int64_t result;
  MPI_Datatype type;
  int *lengths = (int*)calloc(count, sizeof(int));
  MPI_Aint *disps = (MPI_Aint*)calloc(count, sizeof(MPI_Aint));
  for (i=0; i<count; i++) {
    lengths[i] = unhand(blocklengths)->body[i];
    disps[i] = unhand(displacements)->body[i];
  }
  MPI_Type_hindexed(count, lengths, disps, DATATYPE(this), &type);
  REF(result) = type;
  free(lengths);
  free(disps);
  return result;
}

int64_t MPI_Datatype_GetStruct(struct HMPI_Datatype *this,
  long count, HArrayOfInt *blocklengths,
  HArrayOfInt *displacements,
  HArrayOfObject *datatypes) {
  int i;
  int64_t result;
  MPI_Datatype type;
  int *lengths = (int*)calloc(count, sizeof(int));
  MPI_Aint *disps = (MPI_Aint*)calloc(count, sizeof(MPI_Aint));
  MPI_Datatype *types = (MPI_Datatype*)
    calloc(count, sizeof(MPI_Datatype));
  for (i=0; i<count; i++) {
    lengths[i] = unhand(blocklengths)->body[i];
    disps[i] = unhand(displacements)->body[i];
    types[i] = DATATYPE((HMPI_Datatype*)unhand(datatypes)->body[i]);
  }
  MPI_Type_struct(count, lengths, disps, types, &type);
  REF(result) = type;
  free(lengths);
  free(disps);
  free(types);
  return result;
}

