import MPI.*;

class life_mpi {

  static public void main(String[] args) {

    int id, np;
    Status stat;
    String host;

    MPI.Init(args);

    host = MPI.Get_processor_name();
    np = MPI.COMM_WORLD.Size();
    id = MPI.COMM_WORLD.Rank();
    System.out.println("Process "+id+"/"+np+" on "+host);

    int i, j, y, base;

    int N = 11;
    int NITER = 3;

    // Define block

    int[] bsize = new int[1];

    if (N < np) np = N;

    int blockSize = N/np;
    int blockBase = blockSize*id+Math.min(id, N%np);
    if (id < N%np) blockSize++;

    // `block' has `blockSize + 2' columns.  This allows for ghost cells.

    byte block[][] = new byte[blockSize+2][N];
    byte board[][] = (id == 0) ? new byte[N][N] : null;
    byte buffer[]  = (id == 0) ? new byte[N] : null;

    for(i = 0 ; i < blockSize ; i++) {
      int ib = i+1;
      for(y = 0 ; y < N ; y++) {
        int x = blockBase + i ;
        if(x == N / 2 || y == N / 2)
          block [ib] [y] = 1 ;
        else
          block [ib] [y] = 0 ;
      }
    }

    // Main update loop.

    int next = (id+1)%np;
    int prev = (id+np-1)%np;
    int neighbours [][] = new int [blockSize] [N] ;

    for(int iter = 0 ; iter < NITER ; iter++) {

      // Dump state of board to host

      if (id == 0) { // the host

        for (i=0; i < blockSize; i++)
          System.arraycopy(block[i+1], 0, board[i], 0, N);
        base = blockSize;

        for (j=1; j<np; j++) {
          MPI.COMM_WORLD.Recv(new Datatype(bsize), j, 0);
          for (i=0; i<bsize[0]; i++) {
            MPI.COMM_WORLD.Recv(new Datatype(buffer), j, 0);
            System.arraycopy(buffer, 0, board[base+i], 0, N);
          }
          base+=bsize[0];
        }

        System.out.println("Dump Board Status:");
        for (i=0; i<N; i++) {
          for (y=0; y<N; y++)
            System.out.print(board[i][y]+" ");
          System.out.println();
        }

      } else { // the slave

        bsize[0] = blockSize;
        MPI.COMM_WORLD.Send(new Datatype(bsize), 0, 0);
        for(i=0 ; i < blockSize ; i++)
          MPI.COMM_WORLD.Send(new Datatype(block[i+1]), 0, 0);

      }

      // Shift this block's upper edge into next neighbour's lower ghost edge

      MPI.COMM_WORLD.Send(new Datatype(block[blockSize]), next, 0);
      MPI.COMM_WORLD.Recv(new Datatype(block[0]), prev, 0);

      // Shift this block's lower edge into prev neighbour's upper ghost edge

      MPI.COMM_WORLD.Send(new Datatype(block[1]), prev, 0);
      MPI.COMM_WORLD.Recv(new Datatype(block[blockSize+1]), next, 0);

      //  Calculate a block of neighbour sums.

      for(i = 0 ; i < blockSize ; i++) {
        int ib = i + 1 ;
        for(y = 0 ; y < N ; y++) {
          int y_n = (y+N-1) % N ;
          int y_p = (y+1) % N ;
            neighbours [i] [y] =
            block [ib - 1] [y_n] + block [ib - 1] [y] + block [ib - 1] [y_p] +
            block [ib]     [y_n] +                      block [ib]     [y_p] +
            block [ib + 1] [y_n] + block [ib + 1] [y] + block [ib + 1] [y_p] ;
        }
      }

      // Update block of board values.

      for(i = 0 ; i < blockSize ; i++) {
        int ib = i + 1 ;
        for(y = 0 ; y < N ; y++) {
          int neighbour = neighbours [i] [y] ;
          if(neighbour < 2 || neighbour > 3)
            block [ib] [y] = 0 ;
          if(neighbour == 3)
            block [ib] [y] = 1 ;
        }
      }

    }
    MPI.Finalize();
  }
  
}
